#include "zBus.h"

#define SYNC     0x55
#define ESC      0x56

#define IDLE              0
#define WAIT_FOR_SYNC     1
#define WAIT_FOR_ADDRESS  2
#define WAIT_FOR_DATA_LEN 3
#define DATA_READING      4
#define CRC_READING       5

#define DATA(x) ((uint8_t*)(x))

zBus::zBus() {
  state = 0;
  selfId = 0;
  masterSlave = 0;  
}

void zBus::beginMaster(Stream* _stream, int8_t _RxTx) {
	stream = _stream;
  if (_RxTx >= 0) {
    RxTx = _RxTx;
    pinMode(RxTx, HIGH);
    digitalWrite(RxTx, LOW);
  }
}

void zBus::beginSlave (Stream* _stream, uint8_t _selfId, int8_t _RxTx) {
  selfId = _selfId;
  state = WAIT_FOR_SYNC;
  masterSlave = WAIT_FOR_SYNC;
  beginMaster(_stream, _RxTx);
}

#ifdef ESC
  #define sendByte(x)   { if ((x) == SYNC || (x) == ESC) { stream->write(ESC); stream->write(~(x)); } else {stream->write(x); } crc ^= (x); }
#else
  #define sendByte(x)   { stream->write(x); crc ^= (x); }
#endif
#define RxTxEnable(x) { if (RxTx >= 0) digitalWrite(RxTx, x); }

void zBus::sendData(void* _data, uint8_t _size, uint8_t _targetId, uint8_t _request) {
  uint8_t crc = 0;

  RxTxEnable(1);
  stream->write(SYNC);
  if (_targetId) {
    sendByte(_targetId);
  } else {
    sendByte(selfId | 0x80);
  }
  sendByte(_size);
  for(uint8_t i = 0; i < _size; i++){
    sendByte(DATA(_data)[i]);
  }
  sendByte(crc);
  stream->flush();
  RxTxEnable(0);
  
  timeStamp = millis();
  setIdleState(_request);
}

void zBus::sendData(const char* _data, uint8_t _targetId, uint8_t _request) {
  sendData((void*)_data, strlen(_data), _targetId, _request);
}

#ifdef ESC
bool esc = 0;
#endif

boolean zBus::availableData() {
  uint8_t a;
  while (stream->available()) {
    a = stream->read();
//    if (a == SYNC) Serial.println();
//    Serial.print(a, HEX); Serial.print(' ');
#ifdef ESC   
    if (a == ESC) {
      esc = 1;
      continue;
    } else {
      if (esc) {
        a = ~a;
        esc = 0;
      }
    }
#endif    
    switch (state) {
      case WAIT_FOR_SYNC:
        if (a == SYNC) {
          rxIndex = 0;
          state = WAIT_FOR_ADDRESS;
          timeStamp = millis();
          timeOut = TIMEOUT / 4;
        }
        break;
      case WAIT_FOR_ADDRESS:
        if (selfId == 0 || selfId == NO_ADDRESS || selfId == a || a == 0) {
          state = WAIT_FOR_DATA_LEN;
          rxBuffer[rxIndex++] = a;
        } else {
          setIdleState();
        }
        break;
      case WAIT_FOR_DATA_LEN:
        if (a < MAX_LEN) {
          state = DATA_READING;
          rxBuffer[rxIndex++] = a;
        } else {                    // podezrele dlouhy packet
          setIdleState();
        }
        break;
      case DATA_READING:
        rxBuffer[rxIndex++] = a;
        if (rxIndex == rxBuffer[1] + 2) {
          state = CRC_READING;
        }
        break;       
      case CRC_READING:
        uint8_t i, crc = 0;
        for (i = 0; i < rxIndex; i++) {
          crc ^= rxBuffer[i];
        }
        if (crc != a) {
          rxBuffer[0] = NO_ADDRESS;
          rxBuffer[1] = ERROR_CRC;          
        }
        setIdleState();
        return 1;
        break;
    }
  }
  
  if (millis() - timeStamp > timeOut) {
    rxBuffer[0] = NO_ADDRESS;
    rxBuffer[1] = ERROR_TIMEOUT;          
    setIdleState();
    return 1;    
  }
  
  return 0;
}

void zBus::setIdleState(uint8_t _reguest) {
  state = masterSlave;
  if (_reguest) {
    timeOut = TIMEOUT; 
    state = WAIT_FOR_SYNC;
  } else {
    timeOut = 0xFFFFFFFF;
  }
}
uint8_t zBus::getStatus() {
  if (rxBuffer[0] == NO_ADDRESS) {
    return rxBuffer[1];
  }
  return NO_ERROR;
}

void zBus::getData(void* _data, uint8_t _maxSize) {
  _maxSize = min(rxBuffer[1], _maxSize);
  memcpy(_data, &rxBuffer[2], _maxSize);
}

uint8_t zBus::readData(void* _data, uint8_t _maxSize) {
  if (!availableData()) 
    return NO_DATA;

  uint8_t result = getStatus();
  if (result == NO_ERROR) {
    getData(_data, _maxSize);
  }
  return result;
}
